package risc.emulator;

import java.io.FileReader;
import java.io.IOException;

import risc.Opcode;
import risc.RISC;
import risc.CodeParser;

import risc.syscall.Dispatcher;
import risc.syscall.Dispatchee;
import risc.syscall.InputOutput;
import risc.syscall.GarbageCollector;
import risc.syscall.Miscellaneous;
import risc.syscall.Debug;

public class Main {

    //########################################################################
    // Main interface

    public static String usage() {
        return
            "usage: java " +Emulator.class.getName()+ " [options] filename\n" +
            "\n" +
            "where options include:\n" +
            "    -mem=size  set the machine memory size to \"size\"\n" +
			"    -dump      dump memory and registers after execution\n" +
            "    -profile   show profiling statistics after execution\n" +
            "    -profilec  show only total cycles after execution\n" +
			"    -timing=MODEL sets the timing model to one of: \n" +
			"            none     - No pipeline \n" +
			"            IO:w     - In-order \"w\"-way-superscalar pipeline\n" +
			"            OOO:w    - Out-of-order \"w\"-way-superscalar pipeline\n" +
			"    -trace     print an execution trace to stderr.\n" +
			"    -verbose   enable verbose output\n" +
            "    -? -help   print this help message and exit\n";
    }

    private static final int PROFILE_NONE = 0;
    private static final int PROFILE_CYCLES = 1;
    private static final int PROFILE_FULL = 2;
	
    public static int run(String[] args) {
        String filename = null;
        boolean verbose = false;
        boolean dump = false;
        int profile = PROFILE_NONE;
        boolean help = false;
		boolean trace = false;
        int size = Emulator.DEFAULT_MEMORY_SIZE;
		Pipeline pipeline = new OutOfOrderPipeline(2);

        for (int i = 0; i < args.length; i++) 
		{
            if (args[i].equals("-?") || args[i].equals("-help")) {
                help = true;
            } else if (args[i].equals("-dump")) {
                dump = true;
            } else if (args[i].equals("-profilec")) {
                profile = 1;
            } else if (args[i].equals("-profile")) {
                profile = 2;
			} else if (args[i].equals("-trace")) {
				trace = true;
			} else if (args[i].equals("-verbose")) {
                verbose = true;
            } else if (args[i].startsWith("-mem=")) {
                String string = args[i].substring(5);
                try {
                    int value = Integer.parseInt(string);
                    if (value < 0) {
                        System.err.println("invalid memory size: " + value);
                        return 1;
                    } else {
                        size = value;
                    }
                } catch (NumberFormatException exception) {
                    System.err.println("invalid memory size: " + string);
                    return 1;
                }
			} 
			else if (args[i].startsWith("-timing=")) 
			{
				String string = args[i].substring(8);
				if (string.startsWith("OOO:")) 
				{
					String way = string.substring(4);
					try 
					{
						int value = Integer.parseInt(way);
						if (value < 1) 
						{
							System.err.println("invalid number of pipelines: " + value);
							return 1;
						} 
						else 
						{
							pipeline=new OutOfOrderPipeline(value);
						}
					} 
					catch (NumberFormatException exception) 
					{
						System.err.println("invalid pipeline number: " + way);
						return 1;
					}
				}
				else if (string.startsWith("IO:")) 
				{	
					String way = string.substring(3);
					try 
					{
						int value = Integer.parseInt(way);
						if (value < 1) 
						{
							System.err.println("invalid number of pipelines: " + value);
							return 1;
						} 
						else 
						{
							pipeline=new SuperScalarPipeline(value);
						}
					} 
					catch (NumberFormatException exception) 
					{
						System.err.println("invalid pipeline number: " + way);
						return 1;
					}
				}
				else if (string.startsWith("none")) 
				{
							pipeline = new NoneStallingPipeline();
				}
            } 
			else if (args[i].startsWith("-")) 
			{
                System.err.println("unrecognized option: " + args[i]);
                System.err.print(usage());
                return 1;
            } else if (filename == null) {
                filename = args[i];
            } else {
                System.err.println("wrong argument: " + args[i]);
                System.err.print(usage());
                return 1;
            }
        }

        if (help) {
            System.out.print(usage());
            return 0;
        }

        if (filename == null) {
            System.err.print(usage());
            return 1;
        }

        Dispatchee io = new InputOutput();
        Dispatchee gc = new GarbageCollector(verbose);
        Dispatchee db = new Debug();
        Dispatchee misc = new Miscellaneous();
        Dispatcher os = new Dispatcher(new Dispatchee[] {io, gc, db, misc});
        Emulator emulator = new Emulator(size, os, pipeline, trace);

        try {
            emulator.assemble(new FileReader(filename));
        } catch (IOException exception) {
            System.err.println("IO error: " + exception);
            return -1;
        } catch (CodeParser.Exception exception) {
            System.err.println("parse error at " + exception);
            return -1;
        }

        int status = 0;
        try {
            try {
                emulator.run();
            } finally {
                emulator.syscall_nofail(RISC.SYS_IO_FLUSH, 0, 0); // !!!
            }
        } catch (Emulator.Exit exception) {
            status = exception.code;
            if (status != 0)
                System.err.println("exited with status code " + status);
            // !!! duplicated code
        } catch (Emulator.Error exception) {
            System.err.println("runtime error: " + exception.getMessage());
            status = -1;
            dump = true;
        } catch (Emulator.Exception exception) {
            System.err.println("runtime error: " + exception.toString());
            status = -1;
            dump = true;
        }

        if (dump) {
            System.out.println();
            // !!! System.out.print(emulator.dump());
            emulator.syscall_nofail(RISC.SYS_DB_DUMP, 0, 0); // !!!
        }

        if (profile == PROFILE_CYCLES) {
            System.err.println(emulator.time());
        } else if (profile == PROFILE_FULL) {
            System.err.println();
            System.err.println();

            System.err.println("Profiling statistics:");
            System.err.println("Opcode         Count    Stalls  Cycles    Cycles %");
            System.err.println("--------------------------------------------------");

            long[] icntrs = emulator.getInstructionCounters();
			long[] ccntrs = emulator.getCycleCounters();

            long totalCount = 0;
            long totalCycles = emulator.time();
			long totalStalls = 0;

            for (int op = 0; op < RISC.opcode_count; ++op) {
                if (Opcode.codes[op] != null) {
                    totalCount += icntrs[op];
                }
            }

            for (int op = 0; op < RISC.opcode_count; ++op) {
                StringBuffer line = new StringBuffer();
                String opName = RISC.opcode_mnemonic[op];

                if (icntrs[op] > 0) {
					long cycles = ccntrs[op];
					long stalls = emulator.stall(op);
					totalStalls += stalls;

					long percent = 100 * (cycles) / totalCycles;

                    line.append(opName);

                    String countS = String.valueOf(icntrs[op]);
                    padTo(line, 20 - countS.length());
                    line.append(countS);

					String stallsS = String.valueOf(stalls);
					padTo(line, 30 - stallsS.length());
					line.append(stallsS);
					
					String cyclesS = String.valueOf(stalls+cycles);
                    padTo(line, 40 - cyclesS.length());
                    line.append(cyclesS);

					String percentS = String.valueOf(percent);
                    padTo(line, 50 - percentS.length());
                    line.append(percent);

                    System.err.println(line);
                }
            }

            System.err.println("--------------------------------------------------");
            StringBuffer line = new StringBuffer("Total:");

            String totalCountS = String.valueOf(totalCount);
            padTo(line, 20 - totalCountS.length());
            line.append(totalCountS);

            String totalStallsS = String.valueOf(totalStalls);
            padTo(line, 30 - totalStallsS.length());
            line.append(totalStallsS);

			String totalCyclesS = String.valueOf(totalCycles);
			padTo(line, 40 - totalCyclesS.length());
			line.append(totalCyclesS);
		
			System.err.println(line);
			double CPI = ((double)totalCycles)/totalCount;
			System.err.println("CPI: " + String.valueOf(CPI));

        }

        return status;
    }

    private static void padTo(StringBuffer buf, int len) {
        int padLen = len - buf.length();
        for (int i = 0; i < padLen; ++i) buf.append(' ');
    }

    public static void main(String[] args) {
        int status = run(args);
        if (status != 0) System.exit(status);
    }

    //########################################################################
}
