/*
 * SuperScalarPipeline.java
 * 
 * Implements a simple timing model for an in-order N-pipelined Superscalar
 * pipeline.
 * 
 * Up to N instructions can be scheduled in the same cycle as long as there 
 * are no register dependencies on previous instructions.  
 * 
 * An instruction who uses a register r can not be scheduled before the
 * latency of a write to that reigster has passed.
 * 
 * There are no stalls on memory other than the cos of a memory instruction
 * given by the opcode cost.
 * 
 * All branches are predicted correctly and causes no stalls for following
 * instructions.
 * 
 * 
 * The implementation assumes a "global" time (cycle count) that is
 * non-decreasing.  
 * 
 * An array of pipelines keeps track of when each pipeline is ready
 * to accept a new instruction. (If the last instruction (i1) in a pipeline
 * is stalling then a new instruction (i2) can not be scheduled until i1
 * has started executing.
 * 
 * An array of registers keeps track of when the last write to each register
 * will be finished. An instruction that uses a register r can not be scheduled
 * until the last write to that register has finished.
 */
package risc.emulator;

import risc.Opcode;
import risc.RISC;
import risc.CodeWriter;
import risc.StringHelper;


public class SuperScalarPipeline extends Pipeline
{
	/* TODO: replace the pipeline structure with some kind of
	 *       priority queue.
	 */ 
	private int NoPipelines;
	private int[] pipelines;

	public SuperScalarPipeline(int way) 
	{
		this.r = new int[RISC.register_count];
		this.stallCounter =  new long[RISC.opcode_count];
		this.NoPipelines = way;
		this.pipelines = new int[NoPipelines];
	}

	private int find_pipeline() 
	{
		int first = pipelines[0];
		int p = 0;

		// Then try find the first pipeline that finish.
		for(int i = 0; i < NoPipelines; i++) 
		{
			if (pipelines[i] < first) 
			{
				first = pipelines[i];
				p = i;
			}
		}

		return p;
	}

	private int max(int a, int b, int c, int d) 
	{
		int m1 = (a > b) ? a : b;
		int m2 = (m1 > c) ? m1 : c;
		return (m2 > d) ? m2 : d;
	}

	public int schedule(int ir,int time) 
	{
		int opcode = RISC.Helper.op(ir);
		int a = RISC.Helper.ra(ir);
		int b = RISC.Helper.rb(ir);
		int c = RISC.Helper.rc(ir);
		int aTime = 0;
		int bTime = 0;
		int cTime = 0;
		int stall = time;

		// First find out when the instruction can be scheduled,
		// taking dependencies on registers into account.
		int uses = Opcode.codes[opcode].uses;
		if((uses & Opcode.REG_A) > 0) aTime = read(a);
		if((uses & Opcode.REG_B) > 0) bTime = read(b);
		if((uses & Opcode.REG_C) > 0) cTime = read(c);
		if((uses & Opcode.REG_LNK) > 0) cTime = read(31);
		
		// Will this instruction stall?
		stall = max(aTime, bTime, cTime, time);		
		stallCounter[opcode] += stall - (time);

		// Only one instruction can be scheduled per cycle per pipeline
		int p = find_pipeline();
		int p_ready = pipelines[p] < stall ? stall : pipelines[p];
		// Allocate this pipeline
		pipelines[p] = p_ready + 1;

		// Figure out at what cycle the instruction will finish.
		int done = p_ready + Opcode.codes[opcode].cost;

		// Update register stalls for defined registers
		int defs = Opcode.codes[opcode].defs;
		if((defs & Opcode.REG_A) > 0) r[a] = done;
		if((defs & Opcode.REG_B) > 0) r[b] = done;
		if((defs & Opcode.REG_C) > 0) r[c] = done;
		if((defs & Opcode.REG_LNK) > 0) r[31] = done;
		
		return p_ready;
	}
}
