package risc.gui;

/**
 * display a graphical view of the memory usage
 *
 * 05.05.01 by Benjamin Leroy-Beaulieu
 *
 */

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

public class MemoryGUI {

    public JFrame       frame;
    private JLabel       stackLabel;
    private JLabel       freeLabel;
    private JLabel       codeLabel;
    private MemoryCanvas memoryCanvas;

    /** Pointer to the main gui */
    private StartGUI    startGUI;
    
    private class MemoryCanvas extends Canvas {
	private double memorySize, stackSize, codeSize;

	public MemoryCanvas(double memorySize, double stackSize, double codeSize) {
	    super();
	    this.memorySize = memorySize;
	    this.stackSize = stackSize;
	    this.codeSize = codeSize;
	}

	public void setMemorySize(int memorySize) {
	    this.memorySize = memorySize;
	}

	public void setStackSize(int stackSize) {
	    this.stackSize = stackSize;
	}

	public void setCodeSize(int codeSize) {
	    this.codeSize = codeSize;
	}

	public void paint (Graphics g) {
	    Dimension d = getSize();
	    int wCanvas = d.width;
	    int hCanvas = d.height;
	    
	    /** Drawing the rectangle representing the code */
	    int wCode = wCanvas;
	    int hCode = (new Double((hCanvas * codeSize)/memorySize)).intValue();
	    int xCode = 0;
	    int yCode = hCanvas-hCode;
	    g.setColor(Color.red);
	    g.fillRect(xCode, yCode, wCode, hCode);

	    /** Drawing the rectangle representing the memory */
	    int wStack = wCanvas;
	    int hStack = (new Double((hCanvas * stackSize)/memorySize)).intValue();
	    int xStack = 0;
	    int yStack = 0;
	    g.setColor(Color.blue);
	    g.fillRect(xStack, yStack, wStack, hStack);

	    /** Drawing the rectangle representing the free space */
	    int wFree = wCanvas;
	    int hFree = hCanvas - (hCode + hStack);
	    int xFree = 0;
	    int yFree = hStack;
	    g.setColor(Color.white);
	    g.fillRect(xFree, yFree, wFree, hFree);
	} // paint
    } // MemoryCanvas

    public MemoryGUI(StartGUI thisStartGUI) {

	startGUI = thisStartGUI;
	
	/** Creating the frame */
	Toolkit tk  = Toolkit.getDefaultToolkit();
	Dimension d = tk.getScreenSize();
	int screenw = d.width;
	int screenh = d.height;
	frame = new JFrame("Memory");
	frame.setSize(screenw/10, 2*screenh/3);
	frame.setLocation(2*screenw/10,screenh/3);

	frame.setFont(GraphicsEnvironment.getLocalGraphicsEnvironment().getAllFonts()[0]);

	/** Allowing window closing */
        frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

	/** frame's layout Manager */
	frame.getContentPane().setLayout(new GridBagLayout());

	/** frame's constraint object */ 
        GridBagConstraints constraints = new GridBagConstraints();
	// components grow in both directions
        constraints.fill    = GridBagConstraints.BOTH; 
	// 5-pixel margin on all sides
        constraints.insets  = new Insets(5,5,5,5); 
        constraints.weightx = 1.0;
        constraints.weighty = 1.0;

	/**
	 * Calculating the datas
	 */
	int memorySize = startGUI.getEmulator().getMemorySize();
	int stackSize = memorySize - startGUI.getEmulator().getRx(startGUI.SP);
	int codeSize = startGUI.getProgramSize();

	double stackWeight = (double)stackSize / (double)memorySize;
	double codeWeight = (double)codeSize / (double)memorySize;
	double freeWeight = (double)(memorySize - (stackSize+codeSize))/(double)memorySize;
	
	String stackPercentage = (new Double(stackWeight*100)).toString();
	if (stackPercentage.length() > 5)
	    stackPercentage = stackPercentage.substring(0,5);
	String codePercentage = (new Double(codeWeight*100)).toString();
	if (codePercentage.length() > 5)
	    codePercentage = codePercentage.substring(0,5);
	String freePercentage = (new Double(freeWeight*100)).toString();
	if (freePercentage.length() > 5)
	    freePercentage = freePercentage.substring(0,5);
	

	/** MemoryCanvas */
	memoryCanvas = new MemoryCanvas(memorySize,
					stackSize,
					codeSize);
	constraints.gridx = 0;
	constraints.gridy = 0;
	constraints.gridwidth = 1;
	constraints.gridheight = 1;
        constraints.weighty = 1.0;
	frame.getContentPane().add(memoryCanvas, constraints);

	/** stack label */
	stackLabel = new JLabel("Stack: " + stackPercentage + "%");
	constraints.gridx = 0;
	constraints.gridy = 1;
	constraints.gridwidth = 1;
	constraints.gridheight = 1;
        constraints.weighty = 0.0;
	frame.getContentPane().add(stackLabel, constraints);
	stackLabel.setForeground(Color.blue);
	
	/** free label */
	freeLabel = new JLabel("Free: " + freePercentage + "%");
	constraints.gridx = 0;
	constraints.gridy = 2;
	constraints.gridwidth = 1;
	constraints.gridheight = 1;
        constraints.weighty = 0.0;
	frame.getContentPane().add(freeLabel, constraints);
	freeLabel.setForeground(Color.white);
	
	/** code label */
	codeLabel = new JLabel("Code: " + codePercentage + "%");
	constraints.gridx = 0;
	constraints.gridy = 3;
	constraints.gridwidth = 1;
	constraints.gridheight = 1;
        constraints.weighty = 0.0;
	frame.getContentPane().add(codeLabel, constraints);
	codeLabel.setForeground(Color.red);
	
	// !!! frame.setVisible(true);
    } // MemoryGUI constructor;

    protected void updateData() {
	/** MemoryCanvas */
	int memorySize = startGUI.getEmulator().getMemorySize();
	int stackSize = memorySize - startGUI.getEmulator().getRx(startGUI.SP);
	int codeSize = startGUI.getProgramSize();

	memoryCanvas.setCodeSize(codeSize);
	memoryCanvas.setStackSize(stackSize);
	memoryCanvas.repaint();

	/** Labels */
	double stackWeight = (double)stackSize / (double)memorySize;
	double codeWeight = (double)codeSize / (double)memorySize;
	double freeWeight = (double)(memorySize - (stackSize+codeSize))/(double)memorySize;
	String stackPercentage = (new Double(stackWeight*100)).toString();
	if (stackPercentage.length() > 5)
	    stackPercentage = stackPercentage.substring(0,5);
	String codePercentage = (new Double(codeWeight*100)).toString();
	if (codePercentage.length() > 5)
	    codePercentage = codePercentage.substring(0,5);
	String freePercentage = (new Double(freeWeight*100)).toString();
	if (freePercentage.length() > 5)
	    freePercentage = freePercentage.substring(0,5);
	stackLabel.setText("Stack: " + stackPercentage + "%");
	stackLabel.setForeground(Color.blue);
	freeLabel.setText("Free: " + freePercentage + "%");
	freeLabel.setForeground(Color.white);
	codeLabel.setText("Code: " + codePercentage + "%");
	codeLabel.setForeground(Color.red);
    } // updateData

} // MemoryGUI class
	
	
	
	
	
    
