package risc.gui;

import risc.RISC;
import risc.emulator.Emulator;
import risc.emulator.NoneStallingPipeline;
import risc.emulator.OutOfOrderPipeline;
import risc.emulator.Pipeline;
import risc.emulator.SuperScalarPipeline;

public class RiscDebugger {

    private static final int PROFILE_NONE = 0;
    private static final int PROFILE_CYCLES = 1;
    private static final int PROFILE_FULL = 2;

	
    //########################################################################
    // RiscDebugger interface

    public static String usage() {
        return
            "usage: java " +RiscDebugger.class.getName()+ " [options] [filename]\n" +
            "\n" +
            "where options include:\n" +
            "    -mem=size  set the machine memory size to \"size\"\n" +
            "    -verbose   enable verbose output\n" +
            "    -? -help   print this help message and exit\n";
    }

    public static int run(String[] args) {
        String filename = null;
        boolean verbose = false;
        boolean dump = false;
        int profile = PROFILE_NONE;
        boolean help = false;
		boolean trace = false;
        int size = Emulator.DEFAULT_MEMORY_SIZE;
		Pipeline pipeline = new OutOfOrderPipeline(2);

        for (int i = 0; i < args.length; i++) 
		{
            if (args[i].equals("-?") || args[i].equals("-help")) {
                help = true;
            } else if (args[i].equals("-dump")) {
                dump = true;
            } else if (args[i].equals("-profilec")) {
                profile = 1;
            } else if (args[i].equals("-profile")) {
                profile = 2;
			} else if (args[i].equals("-trace")) {
				trace = true;
			} else if (args[i].equals("-verbose")) {
                verbose = true;
            } else if (args[i].startsWith("-mem=")) {
                String string = args[i].substring(5);
                try {
                    int value = Integer.parseInt(string);
                    if (value < 0) {
                        System.err.println("invalid memory size: " + value);
                        return 1;
                    } else {
                        size = value;
                    }
                } catch (NumberFormatException exception) {
                    System.err.println("invalid memory size: " + string);
                    return 1;
                }
			} 
			else if (args[i].startsWith("-timing=")) 
			{
				String string = args[i].substring(8);
				if (string.startsWith("OOO:")) 
				{
					String way = string.substring(4);
					try 
					{
						int value = Integer.parseInt(way);
						if (value < 1) 
						{
							System.err.println("invalid number of pipelines: " + value);
							return 1;
						} 
						else 
						{
							pipeline=new OutOfOrderPipeline(value);
						}
					} 
					catch (NumberFormatException exception) 
					{
						System.err.println("invalid pipeline number: " + way);
						return 1;
					}
				}
				else if (string.startsWith("IO:")) 
				{	
					String way = string.substring(3);
					try 
					{
						int value = Integer.parseInt(way);
						if (value < 1) 
						{
							System.err.println("invalid number of pipelines: " + value);
							return 1;
						} 
						else 
						{
							pipeline=new SuperScalarPipeline(value);
						}
					} 
					catch (NumberFormatException exception) 
					{
						System.err.println("invalid pipeline number: " + way);
						return 1;
					}
				}
				else if (string.startsWith("none")) 
				{
							pipeline = new NoneStallingPipeline();
				}
            } 
			else if (args[i].startsWith("-")) 
			{
                System.err.println("unrecognized option: " + args[i]);
                System.err.print(usage());
                return 1;
            } else if (filename == null) {
                filename = args[i];
            } else {
                System.err.println("wrong argument: " + args[i]);
                System.err.print(usage());
                return 1;
            }
        }

        if (help) {
            System.out.print(usage());
            return 0;
        }
	StartGUI start = new StartGUI(size, verbose, filename, pipeline, trace);

        return 0;
    }

    public static void main(String[] args) {
        int exitcode = run(args);
        if (exitcode != 0) System.exit(exitcode);
    }
    
    //########################################################################
}
