package risc.syscall;

import risc.RISC;
import risc.CodeWriter;
import risc.StringHelper;
import risc.emulator.Emulator;
import risc.emulator.SysCallHandler;

public class Debug extends AbstractHandler implements Dispatchee {

    //########################################################################
    // Debug interface

    // !!! public
    public String toRegisterName(int register) {
        return "R" + register + (register < 10 ? " " : "");
    }

    private String toIntString(int value) {
        return StringHelper.toIntString(value, 11);
    }

    private String toHexString(int value) {
        return "0x" + StringHelper.toHexString(value, 8);
    }

    public String dump(Emulator emu) {
        CodeWriter writer = new CodeWriter();
        StringBuffer buffer = new StringBuffer();

        int pc = emu.getPC() & 0xFFFFFFFC;
        if (pc >= 0 && pc < emu.getMemorySize()) {
            buffer.append("Code:\n");
            for (int adr = Math.max(0, pc - 32); adr <= pc; adr += 4) {
                buffer.append("M[").append(toHexString(adr)).append("] = ");
                int word = emu.getWordAt(adr);
                buffer.append(toHexString(word)).append(" = ");
                writer.appendInstruction(buffer, word);
                buffer.append("\n");
            }
            buffer.append("\n");
        }

        buffer.append("Registers:\n");

        buffer.append("PC ").append(" = ");
        buffer.append(toHexString(emu.getPC())).append(" = ");
        buffer.append(toIntString(emu.getPC())).append("          ");

        buffer.append("IR ").append(" = ");
        buffer.append(toHexString(emu.getIR())).append(" = ");
        writer.appendInstruction(buffer, emu.getIR());
        buffer.append("\n");

        for (int i = 0; i < RISC.register_count; i++) {
            int reg = (RISC.register_count / 2) * (i % 2) + (i / 2);
            buffer.append(toRegisterName(reg)).append(" = ");
            int value = emu.getRx(reg);
            buffer.append(toHexString(value)).append(" = ");
            buffer.append(toIntString(value));
            buffer.append(i % 2 == 0 ? "          " : "\n");
        }
        if (RISC.register_count % 2 != 0) buffer.append("\n");
        buffer.append("\n");

        int sp = emu.getRx(RISC.SP) & 0xFFFFFFFC;
        if (sp >= 0 && sp < emu.getMemorySize()) {
            buffer.append("Stack:\n");
            for (; sp < emu.getMemorySize(); sp += 4) {
                buffer.append("M[").append(toHexString(sp)).append("] = ");
                int word = emu.getWordAt(sp);
                buffer.append(toHexString(word)).append(" = ");
                buffer.append(toIntString(word)).append("\n");
            }
            buffer.append("\n");
        }
        return buffer.toString();
    }

    //########################################################################
    // Dispatchee interface

    public int[] getHandledSysCalls() {
        return new int[] {
            RISC.SYS_DB_WR_REG,
            RISC.SYS_DB_DUMP,
        };
    }

    //########################################################################
    // SysCallHandler interface

    public int handle(int sc, int a, int b, Emulator emu)
        throws Emulator.Exception
    {
        switch (sc) {

        case RISC.SYS_DB_WR_REG:
            write(emu, toRegisterName(RISC.Helper.ra(emu.getIR())));
            write(emu, " = 0x");
            emu.syscall(RISC.SYS_IO_WR_HEX, a, 8);
            write(emu, " = ");
            emu.syscall(RISC.SYS_IO_WR_INT, a, 11);
            write(emu, "");
	    return a;

        case RISC.SYS_DB_DUMP:
            write(emu, dump(emu));
            return a;

        default:
            return super.handle(sc, a, b, emu);
        }
    }

    //########################################################################
    // Private interface

    // !!! throws
    private void write(Emulator emu, String string) throws Emulator.Exception {
        for (int i = 0; i < string.length(); i++) {
            emu.syscall(RISC.SYS_IO_WR_CHR, string.charAt(i), 0);
        }
        emu.syscall(RISC.SYS_IO_FLUSH, 0, 0);
    }

    //########################################################################
}
