package risc.syscall;

import java.io.Writer;
import java.io.PrintWriter;
import java.io.OutputStream;
import java.io.OutputStreamWriter;

import java.util.Iterator;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Stack;
import java.util.BitSet;

import risc.RISC;
import risc.emulator.Emulator;

public class GarbageCollector extends AbstractHandler implements Dispatchee {

    //########################################################################
    // Private state

    private final boolean verbose;
    private final PrintWriter writer;
    private final SortedMap live;
    private final SortedMap dead;
    private final Stack stack;
    private final Cell key;
    private int hp_address;
    private int hp_size;
    private int sp;

    //########################################################################
    // Constructors

    public GarbageCollector() {
        this(false);
    }

    public GarbageCollector(boolean verbose) {
        this(verbose, System.err);
    }

    public GarbageCollector(boolean verbose, OutputStream output) {
        this(verbose, output != null ? new OutputStreamWriter(output) : null);
    }

    public GarbageCollector(boolean verbose, Writer writer) {
        if (verbose && writer == null)
            throw new IllegalArgumentException("null writer");
        if (!(writer instanceof PrintWriter))
            writer = new PrintWriter(writer, true);
        this.verbose = verbose;
        this.writer = (PrintWriter)writer;
        this.live = new TreeMap();
        this.dead = new TreeMap();
        this.stack = new Stack();
        this.key = new Cell(0, 0, false);
        this.hp_address = 0;
        this.hp_size = 0;
        this.sp = 0;
    }

    //########################################################################
    // GarbageCollector interface

    public int getHeapAddress() {
        return hp_address;
    }

    public int getHeapSize() {
        return hp_size;
    }

    public int getHeapUsage() {
        int usage = 0;
        for (Iterator i = live.keySet().iterator(); i.hasNext();) {
            Cell cell = (Cell)i.next();
            usage += cell.size;
        }
        return usage;
    }

    public BitSet getHeapUsageMap() {
        BitSet map = new BitSet(hp_size/4);
        for (Iterator i = live.keySet().iterator(); i.hasNext();) {
            Cell cell = (Cell)i.next();
            int start = (cell.address - hp_address) / 4;
            int count = cell.size / 4;
            for (int j = 0; j < count; j++) map.set(start + j);
        }
        return map;
    }

    //########################################################################
    // Dispatchee interface

    public int[] getHandledSysCalls() {
        return new int[] {
            RISC.SYS_GC_INIT,
            RISC.SYS_GC_ALLOC,
        };
    }

    //########################################################################
    // SysCallHandler interface

    public int handle(int sc, int a, int b, Emulator emu)
        throws Emulator.Exception
    {
        switch (sc) {

        case RISC.SYS_GC_INIT:
            init(a, 4 * (b & 0x01FFFFFF), b >>> 27, emu);
            return a;

        case RISC.SYS_GC_ALLOC:
            return alloc(b, emu);

        default:
            return super.handle(sc, a, b, emu);
        }
    }

    //########################################################################
    // Private interface

    private void init(int hp, int sz, int sp, Emulator emu)
        throws Emulator.Error
    {
        if (hp < 0 || hp >= emu.getMemorySize()) throw error(
            "out of bounds heap start: " + RISC.Helper.getAddressString(hp));
        if (hp + sz - 1 >= emu.getMemorySize()) throw error(
            "out of bounds heap end: " +RISC.Helper.getAddressString(hp+sz-1));
        if (sz < 0) throw error(
            "negative heap size: " + sz);
        if (sp < 0 || sp >= RISC.register_count) throw error(
            "unknown register: " + sp);

        sz = sz == 0 ? emu.getMemorySize() : sz + hp;
        hp = hp + RISC.WORD_SIZE - 1;
        hp = hp - hp % RISC.WORD_SIZE;
        sz = Math.min(sz, emu.getMemorySize());
        sz = Math.max(0, sz - hp);
        sz = sz - sz % RISC.WORD_SIZE;

        if (verbose) {
            writer.print("[GC] ");
            writer.print("heap address = " + hp + ", ");
            writer.print("heap size = " + sz + " Byte, ");
            writer.print(sp == 0 ? "no stack pointer" :
                "stack pointer = " + RISC.register_name[sp]);
            writer.println();
        }

        Cell cell = new Cell(hp, sz, false);
        this.live.clear();
        this.dead.clear();
        this.dead.put(cell, cell);
        this.hp_address = hp;
        this.hp_size = sz;
        this.sp = sp;
    }

    private int alloc(int size, Emulator emu) throws Emulator.Error {
        Cell cell = lock(size);
        if (cell == null) {
            free(emu);
            cell = lock(size);
            if (cell == null) throw error(
                "could not allocate block of " + size + " bytes");
        }
        zero(cell, emu);
        return cell.address;
    }

    private void free(Emulator emu) throws Emulator.Error {
        int usage1 = 0;
        for (Iterator i = live.keySet().iterator(); i.hasNext();) {
            Cell cell = (Cell)i.next();
            cell.live = false;
            usage1 += cell.size;
        }

        int stk_address = sp !=  0 ? emu.getRx(sp) : hp_address + hp_size;
        int mem_size = emu.getMemorySize();
        stk_address = stk_address & 0xFFFFFFFC;
        if (stk_address < hp_address + hp_size) throw error(
            "stack overwrote heap");

        for (int i = 0; i < RISC.register_count; i += 1)
            mark(emu.getRx(i));
        for (int i = 0; i < hp_address; i += 4)
            mark(emu.getWordAt(i));
        for (int i = stk_address; i < mem_size; i += 4)
            mark(emu.getWordAt(i));

        while (!stack.empty()) {
            Cell cell = (Cell)stack.pop();
            for (int i = cell.address; i < cell.address + cell.size; i += 4)
                mark(emu.getWordAt(i));
        }

        int usage2 = 0;
        for (Iterator i = live.keySet().iterator(); i.hasNext(); ) {
            Cell cell = (Cell)i.next();
            if (cell.live) {
                usage2 += cell.size;
            } else {
                i.remove();
                dead.put(cell, cell);
            }
        }

        if (verbose) {
            writer.print("[GC] ");
            writer.print(usage1 + " Byte -> " + usage2 + " Byte ");
            writer.print("(" + hp_size + " Byte)");
            writer.println();
        }
    }

    private void zero(Cell cell, Emulator emu) {
        for (int i = 0; i < cell.size; i += 4) emu.setWordAt(cell.address+i,0);
    }

    private Cell lock(int size) throws Emulator.Error {
        if (size < 0) throw error("negative block size: " + size);
        size = size + RISC.WORD_SIZE - 1;
        size = size - size % RISC.WORD_SIZE;
        size = Math.max(4, size);
        Cell last = null;
        Iterator i = dead.keySet().iterator();
        while (i.hasNext()) {
            Cell cell = (Cell)i.next();
            if (last != null && last.address + last.size == cell.address) {
                i.remove();
                last.size += cell.size;
                cell = last;
            }
            if (cell.size >= size) {
                if (cell.size == size) {
                    cell.live = false;
                    dead.remove(cell);
                } else {
                    // !!!
                    cell.address += size;
                    cell.size -= size;
                    cell = new Cell(cell.address - size, size, true);
                }
                live.put(cell, cell);
                return cell;
            }
            last = cell;
        }
        return null;
    }

    private void mark(int address) {
        key.address = address;
        Cell cell = (Cell)live.get(key);
        if (cell != null && !cell.live) {
            cell.live = true;
            stack.push(cell);
        }
    }

    private Emulator.Error error(String error) {
        return new Emulator.Error("GC error: " + error);
    }

    //########################################################################
    // Cell class

    private static final class Cell implements Comparable {
        public int address;
        public int size;
        public boolean live;

        private Cell(int address, int size, boolean live) {
            this.address = address;
            this.size = size;
            this.live = live;
        }

        public int compareTo(Object object) {
            Cell that = (Cell)object;
            return this.address - that.address;
        }

        public boolean equals(Object object) {
            Cell that = (Cell)object;
            return this.address == that.address;
        }
    }

    //########################################################################
}
