package risc.syscall;

import java.io.Reader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.InputStream;
import java.io.Writer;
import java.io.PrintWriter;
import java.io.OutputStreamWriter;
import java.io.OutputStream;
import java.io.IOException;

import risc.RISC;
import risc.StringHelper; // !!!
import risc.emulator.Emulator;

public class InputOutput extends AbstractHandler implements Dispatchee {

    //########################################################################
    // Private state

    private final BufferedReader reader;
    private final PrintWriter writer;

    //########################################################################
    // Constructors

    public InputOutput() {
        this(System.in, System.out);
    }

    public InputOutput(InputStream input, OutputStream output) {
        this(new InputStreamReader(input), new OutputStreamWriter(output));
    }

    public InputOutput(Reader reader, Writer writer) {
        if (reader == null) throw new IllegalArgumentException("null reader");
        if (writer == null) throw new IllegalArgumentException("null writer");
        this.reader = reader instanceof BufferedReader ?
            (BufferedReader)reader : new BufferedReader(reader);
        this.writer = writer instanceof PrintWriter ?
            (PrintWriter)writer : new PrintWriter(writer);
    }

    //########################################################################
    // InputOutput interface

    public void flush() {
        writer.flush();
    }

    //########################################################################
    // Dispatchee interface

    public int[] getHandledSysCalls() {
        return new int[] {
            RISC.SYS_IO_RD_CHR,
            RISC.SYS_IO_RD_INT,
            RISC.SYS_IO_RD_BIN,
            RISC.SYS_IO_RD_DEC,
            RISC.SYS_IO_RD_HEX,
            RISC.SYS_IO_WR_CHR,
            RISC.SYS_IO_WR_INT,
            RISC.SYS_IO_WR_BIN,
            RISC.SYS_IO_WR_DEC,
            RISC.SYS_IO_WR_HEX,
            RISC.SYS_IO_WR_LN,
            RISC.SYS_IO_FLUSH,
        };
    }

    //########################################################################
    // SysCallHandler interface

    public int handle(int sc, int a, int b, Emulator emu)
        throws Emulator.Exception
    {
        try {
            switch (sc) {

            case RISC.SYS_IO_RD_CHR:
                writer.flush();
                return reader.read();

            case RISC.SYS_IO_RD_INT:
                writer.flush();
		return Integer.parseInt(reader.readLine());

            case RISC.SYS_IO_RD_BIN:
                writer.flush();
                return Integer.parseInt(reader.readLine(), 2);

            case RISC.SYS_IO_RD_DEC:
                writer.flush();
                return Integer.parseInt(reader.readLine(), 10);

            case RISC.SYS_IO_RD_HEX:
                writer.flush();
                return Integer.parseInt(reader.readLine(), 16);

            case RISC.SYS_IO_WR_CHR:
                writer.print((char)a);
                return a;

            case RISC.SYS_IO_WR_INT:
                writer.print(StringHelper.toIntString(a, b));
                return a;

            case RISC.SYS_IO_WR_BIN:
                writer.print(StringHelper.toBinString(a, b));
                return a;

            case RISC.SYS_IO_WR_DEC:
                writer.print(StringHelper.toDecString(a, b));
                return a;

            case RISC.SYS_IO_WR_HEX:
                writer.print(StringHelper.toHexString(a, b));
                return a;

            case RISC.SYS_IO_WR_LN:
                writer.println();
                return a;

            case RISC.SYS_IO_FLUSH:
                writer.flush();
                return a;

            default:
                return super.handle(sc, a, b, emu);
            }
        } catch (IOException exception) {
            throw new Emulator.Error("IO error", exception);
        }
    }

    //########################################################################
}
